#!/usr/bin/env bash
# SPDX-FileCopyrightText: 2023 Fushan Wen <qydwhotmail@gmail.com>
# SPDX-License-Identifier: BSD-2-Clause
$XGETTEXT `find . -name \*.qml` -o $podir/plasma_applet_org.kde.plasma.cameraindicator.pot
