/*
    SPDX-FileCopyrightText: 2024 Fushan Wen <qydwhotmail@gmail.com>

    SPDX-License-Identifier: LGPL-2.1-or-later
*/

#pragma once

#include <QVariant>

namespace Encoder
{
QVariant encode(const QVariant &arg, const char *signature);
}
