/*
    localegeneratorgeneratedglibc.cpp
    SPDX-FileCopyrightText: 2022 Han Young <hanyoung@protonmail.com>

    SPDX-License-Identifier: GPL-2.0-or-later
*/
#include "localegeneratorgeneratedglibc.h"

void LocaleGeneratorGeneratedGlibc::localesGenerate(const QStringList &list)
{
    Q_UNUSED(list)
    Q_EMIT needsFont();
}

#include "moc_localegeneratorgeneratedglibc.cpp"
