/*
    SPDX-License-Identifier: GPL-2.0-or-later
*/

#pragma once

class DayData
{
public:
    bool isCurrent;
    //   bool containsHolidayItems;
    //   bool containsEventItems;
    //   bool containsTodoItems;
    //   bool containsJournalItems;
    int dayNumber;
    int monthNumber;
    int yearNumber;
};
