#pragma once

/*
 * SPDX-FileCopyrightText: 2003-2009 Craig Drummond <craig@kde.org>
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

class QString;

namespace KFI::FcConfig
{
void addDir(const QString &dir, bool system);
}
