# SPDX-FileCopyrightText: 2023 Fushan Wen <qydwhotmail@gmail.com>
# SPDX-License-Identifier: MIT

"""Common constants for Appium tests."""

APPIUM_SERVER_URL: str = "http://127.0.0.1:4723"
KDE_VERSION: int = 6
DEFAULT_IMPLICIT_TIMEOUT: int = 10000
DEFAULT_WAIT_TIMEOUT: int = 10
PROCESS_WAIT_ITERATIONS: int = 10
PROCESS_WAIT_INTERVAL: float = 1.0
