/* UnicodeBlocks.h */
/* THIS IS A GENERATED FILE. CHANGES WILL BE OVERWRITTEN. */
/* Generated by generate-unicode-tables.pl */
/* Generated from UCD version 5.0 */

#pragma once

#include <KLazyLocalizedString>
#include <KLocalizedString>
#include <qglobal.h>

struct TUnicodeBlock {
    quint32 start, end;
    const KLazyLocalizedString blockName;
};

static const struct TUnicodeBlock constUnicodeBlocks[] = {{0x0000, 0x007F, kli18n("Basic Latin")},
                                                          {0x0080, 0x00FF, kli18n("Latin-1 Supplement")},
                                                          {0x0100, 0x017F, kli18n("Latin Extended-A")},
                                                          {0x0180, 0x024F, kli18n("Latin Extended-B")},
                                                          {0x0250, 0x02AF, kli18n("IPA Extensions")},
                                                          {0x02B0, 0x02FF, kli18n("Spacing Modifier Letters")},
                                                          {0x0300, 0x036F, kli18n("Combining Diacritical Marks")},
                                                          {0x0370, 0x03FF, kli18n("Greek and Coptic")},
                                                          {0x0400, 0x04FF, kli18n("Cyrillic")},
                                                          {0x0500, 0x052F, kli18n("Cyrillic Supplement")},
                                                          {0x0530, 0x058F, kli18n("Armenian")},
                                                          {0x0590, 0x05FF, kli18n("Hebrew")},
                                                          {0x0600, 0x06FF, kli18n("Arabic")},
                                                          {0x0700, 0x074F, kli18n("Syriac")},
                                                          {0x0750, 0x077F, kli18n("Arabic Supplement")},
                                                          {0x0780, 0x07BF, kli18n("Thaana")},
                                                          {0x07C0, 0x07FF, kli18n("NKo")},
                                                          {0x0900, 0x097F, kli18n("Devanagari")},
                                                          {0x0980, 0x09FF, kli18n("Bengali")},
                                                          {0x0A00, 0x0A7F, kli18n("Gurmukhi")},
                                                          {0x0A80, 0x0AFF, kli18n("Gujarati")},
                                                          {0x0B00, 0x0B7F, kli18n("Oriya")},
                                                          {0x0B80, 0x0BFF, kli18n("Tamil")},
                                                          {0x0C00, 0x0C7F, kli18n("Telugu")},
                                                          {0x0C80, 0x0CFF, kli18n("Kannada")},
                                                          {0x0D00, 0x0D7F, kli18n("Malayalam")},
                                                          {0x0D80, 0x0DFF, kli18n("Sinhala")},
                                                          {0x0E00, 0x0E7F, kli18n("Thai")},
                                                          {0x0E80, 0x0EFF, kli18n("Lao")},
                                                          {0x0F00, 0x0FFF, kli18n("Tibetan")},
                                                          {0x1000, 0x109F, kli18n("Myanmar")},
                                                          {0x10A0, 0x10FF, kli18n("Georgian")},
                                                          {0x1100, 0x11FF, kli18n("Hangul Jamo")},
                                                          {0x1200, 0x137F, kli18n("Ethiopic")},
                                                          {0x1380, 0x139F, kli18n("Ethiopic Supplement")},
                                                          {0x13A0, 0x13FF, kli18n("Cherokee")},
                                                          {0x1400, 0x167F, kli18n("Unified Canadian Aboriginal Syllabics")},
                                                          {0x1680, 0x169F, kli18n("Ogham")},
                                                          {0x16A0, 0x16FF, kli18n("Runic")},
                                                          {0x1700, 0x171F, kli18n("Tagalog")},
                                                          {0x1720, 0x173F, kli18n("Hanunoo")},
                                                          {0x1740, 0x175F, kli18n("Buhid")},
                                                          {0x1760, 0x177F, kli18n("Tagbanwa")},
                                                          {0x1780, 0x17FF, kli18n("Khmer")},
                                                          {0x1800, 0x18AF, kli18n("Mongolian")},
                                                          {0x1900, 0x194F, kli18n("Limbu")},
                                                          {0x1950, 0x197F, kli18n("Tai Le")},
                                                          {0x1980, 0x19DF, kli18n("New Tai Lue")},
                                                          {0x19E0, 0x19FF, kli18n("Khmer Symbols")},
                                                          {0x1A00, 0x1A1F, kli18n("Buginese")},
                                                          {0x1B00, 0x1B7F, kli18n("Balinese")},
                                                          {0x1D00, 0x1D7F, kli18n("Phonetic Extensions")},
                                                          {0x1D80, 0x1DBF, kli18n("Phonetic Extensions Supplement")},
                                                          {0x1DC0, 0x1DFF, kli18n("Combining Diacritical Marks Supplement")},
                                                          {0x1E00, 0x1EFF, kli18n("Latin Extended Additional")},
                                                          {0x1F00, 0x1FFF, kli18n("Greek Extended")},
                                                          {0x2000, 0x206F, kli18n("General Punctuation")},
                                                          {0x2070, 0x209F, kli18n("Superscripts and Subscripts")},
                                                          {0x20A0, 0x20CF, kli18n("Currency Symbols")},
                                                          {0x20D0, 0x20FF, kli18n("Combining Diacritical Marks for Symbols")},
                                                          {0x2100, 0x214F, kli18n("Letter-Like Symbols")},
                                                          {0x2150, 0x218F, kli18n("Number Forms")},
                                                          {0x2190, 0x21FF, kli18n("Arrows")},
                                                          {0x2200, 0x22FF, kli18n("Mathematical Operators")},
                                                          {0x2300, 0x23FF, kli18n("Miscellaneous Technical")},
                                                          {0x2400, 0x243F, kli18n("Control Pictures")},
                                                          {0x2440, 0x245F, kli18n("Optical Character Recognition")},
                                                          {0x2460, 0x24FF, kli18n("Enclosed Alphanumerics")},
                                                          {0x2500, 0x257F, kli18n("Box Drawing")},
                                                          {0x2580, 0x259F, kli18n("Block Elements")},
                                                          {0x25A0, 0x25FF, kli18n("Geometric Shapes")},
                                                          {0x2600, 0x26FF, kli18n("Miscellaneous Symbols")},
                                                          {0x2700, 0x27BF, kli18n("Dingbats")},
                                                          {0x27C0, 0x27EF, kli18n("Miscellaneous Mathematical Symbols-A")},
                                                          {0x27F0, 0x27FF, kli18n("Supplemental Arrows-A")},
                                                          {0x2800, 0x28FF, kli18n("Braille Patterns")},
                                                          {0x2900, 0x297F, kli18n("Supplemental Arrows-B")},
                                                          {0x2980, 0x29FF, kli18n("Miscellaneous Mathematical Symbols-B")},
                                                          {0x2A00, 0x2AFF, kli18n("Supplemental Mathematical Operators")},
                                                          {0x2B00, 0x2BFF, kli18n("Miscellaneous Symbols and Arrows")},
                                                          {0x2C00, 0x2C5F, kli18n("Glagolitic")},
                                                          {0x2C60, 0x2C7F, kli18n("Latin Extended-C")},
                                                          {0x2C80, 0x2CFF, kli18n("Coptic")},
                                                          {0x2D00, 0x2D2F, kli18n("Georgian Supplement")},
                                                          {0x2D30, 0x2D7F, kli18n("Tifinagh")},
                                                          {0x2D80, 0x2DDF, kli18n("Ethiopic Extended")},
                                                          {0x2E00, 0x2E7F, kli18n("Supplemental Punctuation")},
                                                          {0x2E80, 0x2EFF, kli18n("CJK Radicals Supplement")},
                                                          {0x2F00, 0x2FDF, kli18n("Kangxi Radicals")},
                                                          {0x2FF0, 0x2FFF, kli18n("Ideographic Description Characters")},
                                                          {0x3000, 0x303F, kli18n("CJK Symbols and Punctuation")},
                                                          {0x3040, 0x309F, kli18n("Hiragana")},
                                                          {0x30A0, 0x30FF, kli18n("Katakana")},
                                                          {0x3100, 0x312F, kli18n("Bopomofo")},
                                                          {0x3130, 0x318F, kli18n("Hangul Compatibility Jamo")},
                                                          {0x3190, 0x319F, kli18n("Kanbun")},
                                                          {0x31A0, 0x31BF, kli18n("Bopomofo Extended")},
                                                          {0x31C0, 0x31EF, kli18n("CJK Strokes")},
                                                          {0x31F0, 0x31FF, kli18n("Katakana Phonetic Extensions")},
                                                          {0x3200, 0x32FF, kli18n("Enclosed CJK Letters and Months")},
                                                          {0x3300, 0x33FF, kli18n("CJK Compatibility")},
                                                          {0x3400, 0x4DBF, kli18n("CJK Unified Ideographs Extension A")},
                                                          {0x4DC0, 0x4DFF, kli18n("Yijing Hexagram Symbols")},
                                                          {0x4E00, 0x9FFF, kli18n("CJK Unified Ideographs")},
                                                          {0xA000, 0xA48F, kli18n("Yi Syllables")},
                                                          {0xA490, 0xA4CF, kli18n("Yi Radicals")},
                                                          {0xA700, 0xA71F, kli18n("Modifier Tone Letters")},
                                                          {0xA720, 0xA7FF, kli18n("Latin Extended-D")},
                                                          {0xA800, 0xA82F, kli18n("Syloti Nagri")},
                                                          {0xA840, 0xA87F, kli18n("Phags-pa")},
                                                          {0xAC00, 0xD7AF, kli18n("Hangul Syllables")},
                                                          {0xD800, 0xDB7F, kli18n("High Surrogates")},
                                                          {0xDB80, 0xDBFF, kli18n("High Private Use Surrogates")},
                                                          {0xDC00, 0xDFFF, kli18n("Low Surrogates")},
                                                          {0xE000, 0xF8FF, kli18n("Private Use Area")},
                                                          {0xF900, 0xFAFF, kli18n("CJK Compatibility Ideographs")},
                                                          {0xFB00, 0xFB4F, kli18n("Alphabetic Presentation Forms")},
                                                          {0xFB50, 0xFDFF, kli18n("Arabic Presentation Forms-A")},
                                                          {0xFE00, 0xFE0F, kli18n("Variation Selectors")},
                                                          {0xFE10, 0xFE1F, kli18n("Vertical Forms")},
                                                          {0xFE20, 0xFE2F, kli18n("Combining Half Marks")},
                                                          {0xFE30, 0xFE4F, kli18n("CJK Compatibility Forms")},
                                                          {0xFE50, 0xFE6F, kli18n("Small Form Variants")},
                                                          {0xFE70, 0xFEFF, kli18n("Arabic Presentation Forms-B")},
                                                          {0xFF00, 0xFFEF, kli18n("Half-Width and Full-Width Forms")},
                                                          {0xFFF0, 0xFFFF, kli18n("Specials")},
                                                          {0x10000, 0x1007F, kli18n("Linear B Syllabary")},
                                                          {0x10080, 0x100FF, kli18n("Linear B Ideograms")},
                                                          {0x10100, 0x1013F, kli18n("Aegean Numbers")},
                                                          {0x10140, 0x1018F, kli18n("Ancient Greek Numbers")},
                                                          {0x10300, 0x1032F, kli18n("Old Italic")},
                                                          {0x10330, 0x1034F, kli18n("Gothic")},
                                                          {0x10380, 0x1039F, kli18n("Ugaritic")},
                                                          {0x103A0, 0x103DF, kli18n("Old Persian")},
                                                          {0x10400, 0x1044F, kli18n("Deseret")},
                                                          {0x10450, 0x1047F, kli18n("Shavian")},
                                                          {0x10480, 0x104AF, kli18n("Osmanya")},
                                                          {0x10800, 0x1083F, kli18n("Cypriot Syllabary")},
                                                          {0x10900, 0x1091F, kli18n("Phoenician")},
                                                          {0x10A00, 0x10A5F, kli18n("Kharoshthi")},
                                                          {0x12000, 0x123FF, kli18n("Cuneiform")},
                                                          {0x12400, 0x1247F, kli18n("Cuneiform Numbers and Punctuation")},
                                                          {0x1D000, 0x1D0FF, kli18n("Byzantine Musical Symbols")},
                                                          {0x1D100, 0x1D1FF, kli18n("Musical Symbols")},
                                                          {0x1D200, 0x1D24F, kli18n("Ancient Greek Musical Notation")},
                                                          {0x1D300, 0x1D35F, kli18n("Tai Xuan Jing Symbols")},
                                                          {0x1D360, 0x1D37F, kli18n("Counting Rod Numerals")},
                                                          {0x1D400, 0x1D7FF, kli18n("Mathematical Alphanumeric Symbols")},
                                                          {0x20000, 0x2A6DF, kli18n("CJK Unified Ideographs Extension B")},
                                                          {0x2F800, 0x2FA1F, kli18n("CJK Compatibility Ideographs Supplement")},
                                                          {0xE0000, 0xE007F, kli18n("Tags")},
                                                          {0xE0100, 0xE01EF, kli18n("Variation Selectors Supplement")},
                                                          {0xF0000, 0xFFFFF, kli18n("Supplementary Private Use Area-A")},
                                                          {0x100000, 0x10FFFF, kli18n("Supplementary Private Use Area-B")},
                                                          {0x0, 0x0, {}}};
