/*
 * This file was generated by qdbusxml2cpp version 0.8
 * Command line was: qdbusxml2cpp -m -p FontinstIface org.kde.fontinst.xml -i Family.h
 *
 * qdbusxml2cpp is Copyright (C) 2014 Digia Plc and/or its subsidiary(-ies).
 *
 * This is an auto-generated file.
 * This file may have been hand-edited. Look for HAND-EDIT comments
 * before re-generating it.
 */

#include "FontinstIface.h"

/*
 * Implementation of interface class OrgKdeFontinstInterface
 */

OrgKdeFontinstInterface::OrgKdeFontinstInterface(const QString &service, const QString &path, const QDBusConnection &connection, QObject *parent)
    : QDBusAbstractInterface(service, path, staticInterfaceName(), connection, parent)
{
}

OrgKdeFontinstInterface::~OrgKdeFontinstInterface()
{
}

#include "moc_FontinstIface.cpp"
